@@ -13,8 +13,8 @@
 **  Try to avoid changing these.
 */
 #define RELEASE "1"
-#define PATCHLEVEL "5.1"
-#define DATE "17-Dec-1996"
+#define PATCHLEVEL "7"
+#define DATE "16-Oct-1997"
 
 
 /*
@@ -23,3 +23,6 @@
 **  for example:
 **	INN 1.0 10-Feb-92 (FOO.COM)
 */
+#define RELEASE_C 1
+#define PATCHLEVEL_C 7
+#define SUB_PATCHLEVEL_C 0
